from __future__ import annotations

from palimpzest.constants import MODEL_CARDS, Model, PromptStrategy
from palimpzest.core.data.dataclasses import GenerationStats, OperatorCostEstimates
from palimpzest.core.elements.records import DataRecord
from palimpzest.core.lib.fields import Field
from palimpzest.query.generators.generators import generator_factory
from palimpzest.query.operators.convert import LLMConvert

# TYPE DEFINITIONS
FieldName = str


class MixtureOfAgentsConvert(LLMConvert):

    def __init__(
        self,
        proposer_models: list[Model],
        temperatures: list[float],
        aggregator_model: Model,
        proposer_prompt_strategy: PromptStrategy = PromptStrategy.COT_MOA_PROPOSER,
        aggregator_prompt_strategy: PromptStrategy = PromptStrategy.COT_MOA_AGG,
        proposer_prompt: str | None = None,
        *args,
        **kwargs,
    ):
        kwargs["model"] = None
        kwargs["prompt_strategy"] = None
        super().__init__(*args, **kwargs)
        sorted_proposers, sorted_temps = zip(*[(m, t) for m, t in sorted(zip(proposer_models, temperatures), key=lambda pair: pair[0])])
        self.proposer_models = list(sorted_proposers)
        self.temperatures = list(sorted_temps)
        self.aggregator_model = aggregator_model
        self.proposer_prompt_strategy = proposer_prompt_strategy
        self.aggregator_prompt_strategy = aggregator_prompt_strategy
        self.proposer_prompt = proposer_prompt

        # create generators
        self.proposer_generators = [
            generator_factory(model, self.proposer_prompt_strategy, self.cardinality, self.verbose)
            for model in proposer_models
        ]
        self.aggregator_generator = generator_factory(aggregator_model, self.aggregator_prompt_strategy, self.cardinality, self.verbose)

    def __str__(self):
        op = super().__str__()
        op += f"    Proposer Models: {self.proposer_models}\n"
        op += f"    Temperatures: {self.temperatures}\n"
        op += f"    Aggregator Model: {self.aggregator_model}\n"
        op += f"    Proposer Prompt Strategy: {self.proposer_prompt_strategy.value}\n"
        op += f"    Aggregator Prompt Strategy: {self.aggregator_prompt_strategy.value}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        id_params = {
            "proposer_models": [model.value for model in self.proposer_models],
            "temperatures": self.temperatures,
            "aggregator_model": self.aggregator_model.value,
            "proposer_prompt_strategy": self.proposer_prompt_strategy.value,
            "aggregator_prompt_strategy": self.aggregator_prompt_strategy.value,
            **id_params,
        }

        return id_params

    def get_op_params(self):
        op_params = super().get_op_params()
        op_params = {
            "proposer_models": self.proposer_models,
            "temperatures": self.temperatures,
            "aggregator_model": self.aggregator_model,
            "proposer_prompt_strategy": self.proposer_prompt_strategy,
            "aggregator_prompt_strategy": self.aggregator_prompt_strategy,
            **op_params,
        }

        return op_params

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        """
        Currently, we are using multiple proposer models with different temperatures to synthesize
        answers, which are then aggregated and summarized by a single aggregator model. Thus, we
        roughly expect to incur the cost and time of an LLMConvert * (len(proposer_models) + 1).
        In practice, this naive quality estimate will be overwritten by the CostModel's estimate
        once it executes a few instances of the operator.
        """
        # temporarily set self.model so that super().naive_cost_estimates(...) can compute an estimate
        self.model = self.proposer_models[0]

        # get naive cost estimates for single LLM call and scale it by number of LLMs used in MoA
        naive_op_cost_estimates = super().naive_cost_estimates(source_op_cost_estimates)
        naive_op_cost_estimates.time_per_record *= (len(self.proposer_models) + 1)
        naive_op_cost_estimates.time_per_record_lower_bound = naive_op_cost_estimates.time_per_record
        naive_op_cost_estimates.time_per_record_upper_bound = naive_op_cost_estimates.time_per_record
        naive_op_cost_estimates.cost_per_record *= (len(self.proposer_models) + 1)
        naive_op_cost_estimates.cost_per_record_lower_bound = naive_op_cost_estimates.cost_per_record
        naive_op_cost_estimates.cost_per_record_upper_bound = naive_op_cost_estimates.cost_per_record

        # for naive setting, estimate quality as mean of all model qualities
        model_qualities = [
            MODEL_CARDS[model.value]["overall"] / 100.0
            for model in self.proposer_models + [self.aggregator_model]
        ]
        naive_op_cost_estimates.quality = sum(model_qualities)/(len(self.proposer_models) + 1)
        naive_op_cost_estimates.quality_lower_bound = naive_op_cost_estimates.quality
        naive_op_cost_estimates.quality_upper_bound = naive_op_cost_estimates.quality

        # reset self.model to be None
        self.model = None

        return naive_op_cost_estimates

    def convert(self, candidate: DataRecord, fields: dict[str, Field]) -> tuple[dict[str, list], GenerationStats]:
        # get input fields
        input_fields = self.get_input_fields()

        # execute generator models in sequence
        proposer_model_final_answers, proposer_model_generation_stats = [], []
        for proposer_generator, temperature in zip(self.proposer_generators, self.temperatures):
            gen_kwargs = {"project_cols": input_fields, "output_schema": self.output_schema, "temperature": temperature}
            _, reasoning, generation_stats, _ = proposer_generator(candidate, fields, json_output=False, **gen_kwargs)
            proposer_text = f"REASONING:{reasoning}\n"
            proposer_model_final_answers.append(proposer_text)
            proposer_model_generation_stats.append(generation_stats)

        # call the aggregator
        gen_kwargs = {
            "project_cols": input_fields,
            "output_schema": self.output_schema,
            "model_responses": proposer_model_final_answers,
        }
        field_answers, _, aggregator_gen_stats, _ = self.aggregator_generator(candidate, fields, **gen_kwargs)

        # compute the total generation stats
        generation_stats = sum(proposer_model_generation_stats) + aggregator_gen_stats

        return field_answers, generation_stats



from __future__ import annotations

from abc import ABC, abstractmethod

from palimpzest.core.data.dataclasses import PlanCost
from palimpzest.query.operators.physical import PhysicalOperator
from palimpzest.query.operators.scan import ScanPhysicalOp
from palimpzest.utils.hash_helpers import hash_for_id


class Plan(ABC):
    @abstractmethod
    def compute_plan_id(self) -> str:
        pass

    @abstractmethod
    def __eq__(self, other) -> bool:
        pass

    @abstractmethod
    def __hash__(self) -> int:
        pass

    @abstractmethod
    def __repr__(self) -> str:
        pass

    @abstractmethod
    def __str__(self) -> str:
        pass

    @abstractmethod
    def __getitem__(self, slice) -> tuple:
        pass

    @abstractmethod
    def __iter__(self) -> iter:
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass

class PhysicalPlan(Plan):
    def __init__(self, operators: list[PhysicalOperator], plan_cost: PlanCost | None = None):
        self.operators = operators
        self.plan_cost = plan_cost if plan_cost is not None else PlanCost(cost=0.0, time=0.0, quality=1.0)
        self.plan_id = self.compute_plan_id()

    def compute_plan_id(self) -> str:
        """
        NOTE: This is NOT a universal ID.

        Two different PhysicalPlan instances with the identical lists of operators will have equivalent plan_ids.
        """
        hash_str = str(tuple(op.get_full_op_id() for op in self.operators))
        return hash_for_id(hash_str)

    def __eq__(self, other):
        return isinstance(other, PhysicalPlan) and self.plan_id == other.plan_id

    def __hash__(self):
        return int(self.plan_id, 16)

    def __repr__(self) -> str:
        return str(self)

    def __str__(self):
        start = self.operators[0]
        plan_str = f" 0. {type(start).__name__} -> {start.output_schema.__name__} \n\n"

        for idx, operator in enumerate(self.operators[1:]):
            plan_str += f" {idx+1}. {str(operator)}\n"

        return plan_str

    def __getitem__(self, slice):
        return self.operators[slice]

    def __iter__(self):
        return iter(self.operators)

    def __len__(self):
        return len(self.operators)

    @staticmethod
    def from_ops_and_sub_plan(ops: list[PhysicalOperator], sub_plan: PhysicalPlan, plan_cost: PlanCost) -> PhysicalPlan:
        # create copies of all logical operators
        copy_sub_plan = [op.copy() for op in sub_plan.operators]
        copy_ops = [op.copy() for op in ops]

        # construct full set of operators
        copy_sub_plan.extend(copy_ops)

        # return the PhysicalPlan
        return PhysicalPlan(operators=copy_sub_plan, plan_cost=plan_cost)


class SentinelPlan(Plan):
    def __init__(self, operator_sets: list[list[PhysicalOperator]]):
        # enforce that first operator_set is a scan and that every operator_set has at least one operator
        if len(operator_sets) > 0:
            assert isinstance(operator_sets[0][0], ScanPhysicalOp), "first operator set must be a scan"
            assert all(len(op_set) > 0 for op_set in operator_sets), "every operator set must have at least one operator"

        # store operator_sets and logical_op_ids; sort operator_sets internally by full_op_id
        self.operator_sets = operator_sets
        self.operator_sets = [sorted(op_set, key=lambda op: op.get_full_op_id()) for op_set in self.operator_sets]
        self.logical_op_ids = [op_set[0].logical_op_id for op_set in self.operator_sets]
        self.plan_id = self.compute_plan_id()

    def compute_plan_id(self) -> str:
        """
        NOTE: This is NOT a universal ID.

        Two different SentinelPlan instances with the identical operator_sets will have equivalent plan_ids.
        """
        hash_str = ""
        for logical_op_id, op_set in zip(self.logical_op_ids, self.operator_sets):
            hash_str += f"{logical_op_id} {tuple(op.get_full_op_id() for op in op_set)} "
        return hash_for_id(hash_str)

    def __eq__(self, other):
        return isinstance(other, SentinelPlan) and self.plan_id == other.plan_id

    def __hash__(self):
        return int(self.plan_id, 16)

    def __repr__(self) -> str:
        return str(self)

    def __str__(self):
        # by assertion, first operator_set is guaranteed to be a scan
        start = self.operator_sets[0][0]
        plan_str = f" 0. {type(start).__name__} -> {start.output_schema.__name__} \n\n"

        # build string one operator set at a time
        for idx, operator_set in enumerate(self.operator_sets[1:]):
            if len(operator_set) == 1:
                operator = operator_set[0]
                plan_str += f" {idx+1}. {str(operator)}\n"

            else:
                for inner_idx, operator in enumerate(operator_set):
                    plan_str += f" {idx+1}.{inner_idx+1}. {str(operator)}\n"

        return plan_str

    def __getitem__(self, slice):
        return self.logical_op_ids[slice], self.operator_sets[slice]

    def __iter__(self):
        yield from zip(self.logical_op_ids, self.operator_sets)

    def __len__(self):
        return len(self.logical_op_ids)

    @staticmethod
    def from_ops_and_sub_plan(op_sets: list[list[PhysicalOperator]], sub_plan: SentinelPlan) -> SentinelPlan:
        # create copies of all logical operators
        copy_sub_plan = [[op.copy() for op in op_set] for op_set in sub_plan.operator_sets]
        copy_ops = [[op.copy() for op in op_set] for op_set in op_sets]

        # construct full set of operators
        copy_sub_plan.extend(copy_ops)

        # return the SentinelPlan
        return SentinelPlan(operator_sets=copy_sub_plan)



import hashlib
import json

from palimpzest.constants import MAX_ID_CHARS


def hash_for_id(id_str: str, max_chars: int = MAX_ID_CHARS) -> str:
    return hashlib.sha256(id_str.encode("utf-8")).hexdigest()[:max_chars]


def hash_for_serialized_dict(dict_obj: dict) -> str:
    return hash_for_id(json.dumps(dict_obj, sort_keys=True))



from enum import Enum

from palimpzest.query.execution.execution_strategy_type import ExecutionStrategyType
from palimpzest.query.processor.nosentinel_processor import NoSentinelQueryProcessor
from palimpzest.query.processor.sentinel_processor import SentinelQueryProcessor
from palimpzest.query.processor.streaming_processor import StreamingQueryProcessor


class ProcessingStrategyType(Enum):
    """How to generate and optimize query plans"""
    SENTINEL = SentinelQueryProcessor
    NO_SENTINEL = NoSentinelQueryProcessor
    STREAMING = StreamingQueryProcessor

    def valid_execution_strategies(self) -> list[ExecutionStrategyType]:
        """
        Returns a list of valid execution strategies for the given processing strategy.
        """
        if self == ProcessingStrategyType.SENTINEL or self == ProcessingStrategyType.NO_SENTINEL:
            return [ExecutionStrategyType.SEQUENTIAL, ExecutionStrategyType.PIPELINED, ExecutionStrategyType.PARALLEL, ExecutionStrategyType.SEQUENTIAL_PARALLEL]
        elif self == ProcessingStrategyType.STREAMING:
            return [ExecutionStrategyType.PIPELINED, ExecutionStrategyType.PARALLEL, ExecutionStrategyType.SEQUENTIAL_PARALLEL]

    def is_sentinel_strategy(self) -> bool:
        """
        Returns True if the query processor associated with this strategy uses sentinel execution.
        """
        return self == ProcessingStrategyType.SENTINEL



from __future__ import annotations

import json

from palimpzest.core.data.dataclasses import OperatorCostEstimates
from palimpzest.core.elements.records import DataRecord, DataRecordSet
from palimpzest.core.lib.schemas import Schema
from palimpzest.utils.hash_helpers import hash_for_id


class PhysicalOperator:
    """
    All implemented physical operators should inherit from this class.
    In order for the Optimizer to consider using a physical operator for a
    given logical operation, the user must also write an ImplementationRule.
    """

    def __init__(
        self,
        output_schema: Schema,
        input_schema: Schema | None = None,
        depends_on: list[str] | None = None,
        logical_op_id: str | None = None,
        logical_op_name: str | None = None,
        target_cache_id: str | None = None,
        verbose: bool = False,
        *args,
        **kwargs,
    ) -> None:
        self.output_schema = output_schema
        self.input_schema = input_schema
        self.depends_on = depends_on if depends_on is None else sorted(depends_on)
        self.logical_op_id = logical_op_id
        self.logical_op_name = logical_op_name
        self.target_cache_id = target_cache_id
        self.verbose = verbose
        self.op_id = None

        # compute the fields generated by this physical operator
        input_field_names = self.input_schema.field_names() if self.input_schema is not None else []
        self.generated_fields = sorted([
            field_name
            for field_name in self.output_schema.field_names()
            if field_name not in input_field_names
        ])

        # sets __hash__() for each child Operator to be the base class' __hash__() method;
        # by default, if a subclass defines __eq__() but not __hash__() Python will set that
        # class' __hash__ to None
        self.__class__.__hash__ = PhysicalOperator.__hash__

    def __str__(self):
        op = f"{self.input_schema.class_name()} -> {self.op_name()} -> {self.output_schema.class_name()}\n"
        op += f"    ({', '.join(self.input_schema.field_names())[:30]}) "
        op += f"-> ({', '.join(self.output_schema.field_names())[:30]})\n"
        if getattr(self, "model", None):
            op += f"    Model: {self.model}\n"
        return op

    def __eq__(self, other) -> bool:
        all_op_params_match = all(value == getattr(other, key) for key, value in self.get_op_params().items())
        return isinstance(other, self.__class__) and all_op_params_match

    def copy(self) -> PhysicalOperator:
        return self.__class__(**self.get_op_params())

    def op_name(self) -> str:
        """Name of the physical operator."""
        return str(self.__class__.__name__)

    def get_id_params(self) -> dict:
        """
        Returns a dictionary mapping of physical operator parameters which are relevant
        for computing the physical operator id.

        NOTE: Should be overriden by subclasses to include class-specific parameters.
        NOTE: input_schema and output_schema are not included in the id params by default,
              because they may depend on the order of operations chosen by the Optimizer.
              This is particularly true for convert operations, where the output schema
              is now the union of the input and output schemas of the logical operator.
        """
        # return {"generated_fields": self.generated_fields}
        return {}

    def get_op_params(self) -> dict:
        """
        Returns a dictionary mapping of physical operator parameters which may be used to
        create a copy of this physical operation.

        NOTE: Should be overriden by subclasses to include class-specific parameters.
        """
        return {
            "output_schema": self.output_schema,
            "input_schema": self.input_schema,
            "depends_on": self.depends_on,
            "logical_op_id": self.logical_op_id,
            "logical_op_name": self.logical_op_name,
            "target_cache_id": self.target_cache_id,
            "verbose": self.verbose,
        }

    def get_op_id(self):
        """
        NOTE: We do not call this in the __init__() method as subclasses may set parameters
              returned by self.get_id_params() after they call to super().__init__().

        NOTE: This is NOT a universal ID.

        Two different PhysicalOperator instances with the identical returned values
        from the call to self.get_id_params() will have equivalent op_ids.
        """
        # return self.op_id if we've computed it before
        if self.op_id is not None:
            return self.op_id

        # get op name and op parameters which are relevant for computing the id
        op_name = self.op_name()
        id_params = self.get_id_params()
        id_params = {
            k: str(v) if k != "output_schema" else sorted(v.field_names())
            for k, v in id_params.items()
        }

        # compute, set, and return the op_id
        hash_str = json.dumps({"op_name": op_name, **id_params}, sort_keys=True)
        self.op_id = hash_for_id(hash_str)

        return self.op_id
    
    def get_logical_op_id(self) -> str | None:
        return self.logical_op_id

    def get_full_op_id(self):
        return f"{self.get_logical_op_id()}-{self.get_op_id()}"

    def __hash__(self):
        return int(self.op_id, 16) # NOTE: should we use self.get_full_op_id() instead?

    def get_model_name(self) -> str | None:
        """Returns the name of the model used by the physical operator (if it sets self.model). Otherwise, it returns None."""
        return None

    def get_input_fields(self):
        """Returns the set of input fields needed to execute a physical operator."""
        depends_on_fields = (
            [field.split(".")[-1] for field in self.depends_on]
            if self.depends_on is not None and len(self.depends_on) > 0
            else None
        )
        input_fields = (
            self.input_schema.field_names()
            if depends_on_fields is None
            else [field for field in self.input_schema.field_names() if field in depends_on_fields]
        )

        return input_fields

    def get_fields_to_generate(self, candidate: DataRecord) -> list[str]:
        """
        Returns the list of field names that this operator needs to generate for the given candidate.
        This function returns only the fields in self.generated_fields which are not already present
        in the candidate. This is important for operators with retry logic, where we may only need to
        recompute a subset of self.generated_fields.

        Right now this is only used by convert and retrieve operators.
        """
        fields_to_generate = [
            field_name
            for field_name in self.generated_fields
            if getattr(candidate, field_name, None) is None
        ]

        return fields_to_generate

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        """
        This function returns a naive estimate of this operator's:
        - cardinality
        - time_per_record
        - cost_per_record
        - quality

        The function takes an argument which contains the OperatorCostEstimates
        of the physical operator whose output is the input to this operator.

        For the implemented operator. These will be used by the CostModel
        when PZ does not have sample execution data -- and it will be necessary
        in some cases even when sample execution data is present. (For example,
        the cardinality of each operator cannot be estimated based on sample
        execution data alone -- thus ScanPhysicalOps need to give at least ballpark
        correct estimates of this quantity).
        """
        raise NotImplementedError("CostEstimates from abstract method")

    def __call__(self, candidate: DataRecord) -> DataRecordSet:
        raise NotImplementedError("Calling __call__ from abstract method")



from __future__ import annotations

import time
from abc import ABC, abstractmethod
from typing import Any

from palimpzest.constants import (
    MODEL_CARDS,
    NAIVE_EST_FILTER_SELECTIVITY,
    NAIVE_EST_NUM_INPUT_TOKENS,
    Cardinality,
    Model,
    PromptStrategy,
)
from palimpzest.core.data.dataclasses import GenerationStats, OperatorCostEstimates, RecordOpStats
from palimpzest.core.elements.filters import Filter
from palimpzest.core.elements.records import DataRecord, DataRecordSet
from palimpzest.core.lib.fields import BooleanField
from palimpzest.query.generators.generators import generator_factory
from palimpzest.query.operators.physical import PhysicalOperator
from palimpzest.utils.model_helpers import get_vision_models


class FilterOp(PhysicalOperator, ABC):
    def __init__(self, filter: Filter, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert self.input_schema.get_desc() == self.output_schema.get_desc(), "Input and output schemas must match for FilterOp"
        self.filter_obj = filter

    def __str__(self):
        op = super().__str__()
        op += f"    Filter: {str(self.filter_obj)}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        return {"filter": str(self.filter_obj), **id_params}

    def get_op_params(self):
        op_params = super().get_op_params()
        return {"filter": self.filter_obj, **op_params}

    @abstractmethod
    def is_image_filter(self) -> bool:
        """Return True if the filter operation processes an image, False otherwise."""
        pass

    @abstractmethod
    def filter(self, candidate: DataRecord) -> tuple[dict[str, bool], GenerationStats]:
        """
        This abstract method will be implemented by subclasses of FilterOp to process the input DataRecord
        and generate the True / False determination of whether the input record passes the filter. A dictionary
        mapping a "passed_operator" key to the T/F boolean is returned along with the GenerationStats object.

        For example, if the input DataRecord (i.e. `candidate`) contains an image of a dog, and the filter
        operation is supposed to filter for images with dogs, then the output would be:

        ({"passed_operator": True}, GenerationStats(...))

        A post-condition of this method is that the "passed_operator" key must be present in the output dictionary,
        and it's value must be a boolean. If there is an error, then the value for "passed_operator" must be False.
        """
        pass

    def _create_record_set(
        self,
        candidate: DataRecord,
        passed_operator: bool,
        generation_stats: GenerationStats,
        total_time: float,
        answer: dict[str, Any],
    ) -> DataRecordSet:
        """
        Given an input DataRecord and a determination of whether it passed the filter or not,
        construct the resulting RecordSet.
        """
        # create new DataRecord and set passed_operator attribute
        dr = DataRecord.from_parent(candidate.schema, parent_record=candidate)
        dr.passed_operator = passed_operator

        # create RecordOpStats object
        record_op_stats = RecordOpStats(
            record_id=dr.id,
            record_parent_id=dr.parent_id,
            record_source_idx=dr.source_idx,
            record_state=dr.to_dict(include_bytes=False),
            full_op_id=self.get_full_op_id(),
            logical_op_id=self.logical_op_id,
            op_name=self.op_name(),
            time_per_record=total_time,
            cost_per_record=generation_stats.cost_per_record,
            model_name=self.get_model_name(),
            filter_str=self.filter_obj.get_filter_str(),
            total_input_tokens=generation_stats.total_input_tokens,
            total_output_tokens=generation_stats.total_output_tokens,
            total_input_cost=generation_stats.total_input_cost,
            total_output_cost=generation_stats.total_output_cost,
            llm_call_duration_secs=generation_stats.llm_call_duration_secs,
            fn_call_duration_secs=generation_stats.fn_call_duration_secs,
            total_llm_calls=generation_stats.total_llm_calls,
            total_embedding_llm_calls=generation_stats.total_embedding_llm_calls,
            answer=answer,
            passed_operator=passed_operator,
            image_operation=self.is_image_filter(),
            op_details={k: str(v) for k, v in self.get_id_params().items()},
        )

        return DataRecordSet([dr], [record_op_stats])

    def __call__(self, candidate: DataRecord) -> DataRecordSet:
        start_time = time.time()

        # apply the filter operation
        field_answers, generation_stats = self.filter(candidate)

        # create and return record set
        record_set = self._create_record_set(
            candidate,
            field_answers["passed_operator"],
            generation_stats,
            time.time() - start_time,
            field_answers
        )

        return record_set


class NonLLMFilter(FilterOp):
    def is_image_filter(self) -> bool:
        # NOTE: even if the UDF is processing an image, we do not consider this an image filter
        # (the output of this function will be used by the CostModel in a way which does not apply to UDFs)
        return False

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates):
        # estimate output cardinality using a constant assumption of the filter selectivity
        selectivity = NAIVE_EST_FILTER_SELECTIVITY
        cardinality = selectivity * source_op_cost_estimates.cardinality

        # estimate 1 ms single-threaded execution for filter function
        time_per_record = 0.001

        # assume filter fn has perfect quality
        return OperatorCostEstimates(
            cardinality=cardinality,
            time_per_record=time_per_record,
            cost_per_record=0.0,
            quality=1.0,
        )
    
    def filter(self, candidate: DataRecord) -> tuple[dict[str, bool], GenerationStats]:
        # apply filter function to input record
        start_time = time.time()
        answer = {}
        try:
            # execute the UDF filter
            passed_operator = self.filter_obj.filter_fn(candidate.to_dict())
            answer = {"passed_operator": passed_operator}

            if self.verbose:
                print(f"{self.filter_obj.get_filter_str()}:\n{passed_operator}")

        except Exception as e:
            print(f"Error invoking user-defined function for filter: {e}")
            raise e

        # create generation stats object containing the time spent executing the UDF function
        generation_stats = GenerationStats(fn_call_duration_secs=time.time() - start_time)

        return answer, generation_stats


class LLMFilter(FilterOp):
    def __init__(
        self,
        model: Model,
        prompt_strategy: PromptStrategy = PromptStrategy.COT_BOOL,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.model = model
        self.prompt_strategy = prompt_strategy
        self.generator = generator_factory(model, prompt_strategy, Cardinality.ONE_TO_ONE, self.verbose)

    def get_id_params(self):
        id_params = super().get_id_params()
        id_params = {
            "model": self.model.value,
            "prompt_strategy": self.prompt_strategy.value,
            **id_params,
        }

        return id_params

    def get_op_params(self):
        op_params = super().get_op_params()
        op_params = {
            "model": self.model,
            "prompt_strategy": self.prompt_strategy,
            **op_params,
        }

        return op_params

    def get_model_name(self):
        return self.model.value

    def is_image_filter(self) -> bool:
        return self.model in get_vision_models()

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates):
        # estimate number of input tokens from source
        est_num_input_tokens = NAIVE_EST_NUM_INPUT_TOKENS
        if self.is_image_filter():
            est_num_input_tokens = 765 / 10  # 1024x1024 image is 765 tokens

        # NOTE: the output often generates an entire reasoning sentence, thus the true value may be higher
        # the filter operation's LLM call should only output TRUE or FALSE, thus we expect its
        # number of output tokens to be ~1.25
        est_num_output_tokens = 1.25

        # get est. of conversion time per record from model card;
        model_conversion_time_per_record = (
            MODEL_CARDS[self.model.value]["seconds_per_output_token"] * est_num_output_tokens
        )

        # get est. of conversion cost (in USD) per record from model card
        model_conversion_usd_per_record = (
            MODEL_CARDS[self.model.value]["usd_per_input_token"] * est_num_input_tokens
            + MODEL_CARDS[self.model.value]["usd_per_output_token"] * est_num_output_tokens
        )

        # estimate output cardinality using a constant assumption of the filter selectivity
        selectivity = NAIVE_EST_FILTER_SELECTIVITY
        cardinality = selectivity * source_op_cost_estimates.cardinality

        # estimate quality of output based on the strength of the model being used
        quality = (MODEL_CARDS[self.model.value]["overall"] / 100.0) * source_op_cost_estimates.quality

        return OperatorCostEstimates(
            cardinality=cardinality,
            time_per_record=model_conversion_time_per_record,
            cost_per_record=model_conversion_usd_per_record,
            quality=quality,
        )

    def filter(self, candidate: DataRecord) -> tuple[dict[str, bool], GenerationStats]:
        # get the set of input fields to use for the filter operation
        input_fields = self.get_input_fields()

        # construct kwargs for generation
        gen_kwargs = {"project_cols": input_fields, "filter_condition": self.filter_obj.filter_condition}

        # generate output; NOTE: BooleanField is used to indicate the output type; thus, the desc is not needed
        fields = {"passed_operator": BooleanField(desc="")}
        field_answers, _, generation_stats, _ = self.generator(candidate, fields, **gen_kwargs)

        return field_answers, generation_stats



from __future__ import annotations

import logging
from typing import Any

from palimpzest.core.data.dataclasses import PlanCost
from palimpzest.policy import Policy
from palimpzest.query.optimizer.cost_model import BaseCostModel
from palimpzest.query.optimizer.optimizer_strategy_type import OptimizationStrategyType
from palimpzest.query.optimizer.primitives import Expression, Group
from palimpzest.query.optimizer.rules import ImplementationRule, Rule, TransformationRule

logger = logging.getLogger(__name__)

class Task:
    """
    Base class for a task. Each task has a method called perform() which executes the task.
    Examples of tasks include optimizing and exploring groups, optimizing expressions, applying
    rules, and optimizing inputs / costing the full group tree.
    """

    def perform(self, groups: dict[int, Group], context: dict[str, Any] | None = None) -> list[Task]:
        """
        NOTE: At the moment we do not make use of the context, but in the future
        this can be used to store required physical properties (e.g. sort conditions
        for the query) and bounds (e.g. the operator should not cost more than X).
        """
        raise NotImplementedError("Calling this method from an abstract base class.")


class OptimizeGroup(Task):
    """
    The task to optimize a group.

    This task pushes optimization tasks for the group's current logical and physical
    expressions onto the tasks stack. This will fully expand the space of possible
    logical and physical expressions for the group, because OptimizeLogicalExpression
    and OptimizePhysicalExpression tasks will indirectly schedule new tasks to apply
    rules and to optimize input groups and expressions.
    """

    def __init__(self, group_id: int):
        self.group_id = group_id

    def perform(self, groups: dict[int, Group], context: dict[str, Any] | None = None) -> list[Task]:
        logger.debug(f"Optimizing group {self.group_id}")
        # get updated instance of the group to be optimized
        if context is None:
            context = {}
        group = groups[self.group_id]

        # if this group has already been optimized, there's nothing more to do
        if group.optimized:
            return []

        # otherwise, optimize all the logical expressions for the group
        new_tasks = []
        for logical_expr in group.logical_expressions:
            task = OptimizeLogicalExpression(logical_expr)
            new_tasks.append(task)

        # and optimize all of the physical expressions in the group
        for physical_expr in group.physical_expressions:
            task = OptimizePhysicalExpression(physical_expr)
            new_tasks.append(task)

        logger.debug(f"Done optimizing group {self.group_id}")
        logger.debug(f"New tasks: {len(new_tasks)}")
        return new_tasks


class ExpandGroup(Task):
    """
    The task to expand a group.

    NOTE: we currently do not use this task, but I'm keeping it around in case we need it
    once we add join operations.
    """

    def __init__(self, group_id: int):
        self.group_id = group_id

    def perform(self, groups: dict[int, Group], context: dict[str, Any] | None = None) -> list[Task]:
        logger.debug(f"Expanding group {self.group_id}")

        # fetch group
        if context is None:
            context = {}
        group = groups[self.group_id]

        # if the group has been explored before, return []
        if group.explored:
            return []

        # for each logical_expr in the group, add a new OptimizeLogicalExpression() task to the queue
        new_tasks = []
        for logical_expr in group.logical_expressions:
            task = OptimizeLogicalExpression(logical_expr, exploring=True)
            new_tasks.append(task)

        # mark the group as explored and return tasks
        group.set_explored()

        logger.debug(f"Done expanding group {self.group_id}")
        logger.debug(f"New tasks: {len(new_tasks)}")
        return new_tasks


class OptimizeLogicalExpression(Task):
    """
    The task to optimize a (multi-)expression.

    This task filters for the subset of rules which may be applied to the given logical expression
    and schedules ApplyRule tasks for each rule.
    """

    def __init__(self, logical_expression: Expression, exploring: bool = False):
        self.logical_expression = logical_expression
        self.exploring = exploring

    def perform(
        self,
        transformation_rules: list[TransformationRule],
        implementation_rules: list[ImplementationRule],
        context: dict[str, Any] | None = None,
    ) -> list[Task]:
        logger.debug(f"Optimizing logical expression {self.logical_expression}")
        # if we're exploring, only apply transformation rules
        if context is None:
            context = {}
        rules = transformation_rules if self.exploring else transformation_rules + implementation_rules

        # filter out rules that have already been applied to logical expression
        rules = list(filter(lambda rule: rule.get_rule_id() not in self.logical_expression.rules_applied, rules))

        # filter for rules that match on this logical expression
        rules = list(filter(lambda rule: rule.matches_pattern(self.logical_expression), rules))

        # TODO compute priority (i.e. "promise") of the rules and sort in order of priority

        # apply rules, exploring the input group(s) of each pattern if necessary
        new_tasks = []
        for rule in rules:
            # TODO: if necessary, expand the input groups of the logical expression to see if they need to be expanded
            apply_rule_task = ApplyRule(rule, self.logical_expression, self.exploring)
            new_tasks.append(apply_rule_task)

        logger.debug(f"Done optimizing logical expression {self.logical_expression}")
        logger.debug(f"New tasks: {len(new_tasks)}")
        return new_tasks


class ApplyRule(Task):
    """
    The task to apply a transformation or implementation rule to a (multi-)expression.

    For TransformationRules, this task will:
    - apply the substitution, receiving new expressions and groups
    - filter the new expressions for ones which may already exist
      - NOTE: we don't filter new groups because this implicitly must be done by the
              transformation rule in order to assign the correct group_id to any
              new expressions.
    - add new expressions to their group's set of logical expressions
    - schedule OptimizeGroup and OptimizeLogicalExpression tasks

    For ImplementationRules, this task will:
    - apply the substitution, receiving new expressions and groups
    - filter the new expressions for ones which may already exist
    - add new expressions to their group's set of physical expressions
    - schedule OptimizePhysicalExpression tasks
    """

    def __init__(self, rule: Rule, logical_expression: Expression, exploring: bool = False):
        self.rule = rule
        self.logical_expression = logical_expression
        self.exploring = exploring

    def perform(
        self,
        groups: dict[int, Group],
        expressions: dict[int, Expression],
        context: dict[str, Any] | None = None,
        **physical_op_params,
    ) -> tuple[list[Task], int]:
        logger.debug(f"Applying rule {self.rule} to logical expression {self.logical_expression}")
        
        # check if rule has already been applied to this logical expression; return [] if so
        if context is None:
            context = {}
        if self.rule.get_rule_id() in self.logical_expression.rules_applied:
            return []

        # MAYBE ?TODO?: iterate over bindings for logical expression and rule?
        #               perhaps some rules can be applied more than once to an expression?

        # get the group of the logical expression
        group_id = self.logical_expression.group_id
        group = groups[group_id]

        # process new expressions; update groups and create new tasks as needed
        new_tasks = []
        if issubclass(self.rule, TransformationRule):
            # apply transformation rule
            new_expressions, new_groups = self.rule.substitute(
                self.logical_expression, groups, expressions, **physical_op_params
            )

            # filter out any expressions which are duplicates (i.e. they've been previously computed)
            new_expressions = [expr for expr in new_expressions if expr.get_expr_id() not in expressions]
            expressions.update({expr.get_expr_id(): expr for expr in new_expressions})

            # add all new groups to the groups mapping
            for group in new_groups:
                groups[group.group_id] = group
                task = OptimizeGroup(group.group_id)

            # add new expressions to their respective groups
            for expr in new_expressions:
                group = groups[expr.group_id]
                group.logical_expressions.add(expr)

            # NOTE: we place new tasks for groups on the top of the stack so that they may be
            #       optimized before we optimize expressions which take new groups as inputs
            # create new tasks for optimizing new logical expressions
            for expr in new_expressions:
                task = OptimizeLogicalExpression(expr, self.exploring)
                new_tasks.append(task)

            # create new tasks for optimizing new groups
            for group in new_groups:
                task = OptimizeGroup(group.group_id)
                new_tasks.append(task)

        else:
            # apply implementation rule
            new_expressions = self.rule.substitute(self.logical_expression, **physical_op_params)
            new_expressions = [expr for expr in new_expressions if expr.get_expr_id() not in expressions]
            costed_full_op_ids = context['costed_full_op_ids']
            if costed_full_op_ids is not None:
                new_expressions = [expr for expr in new_expressions if expr.operator.get_full_op_id() in costed_full_op_ids]
            expressions.update({expr.get_expr_id(): expr for expr in new_expressions})
            group.physical_expressions.update(new_expressions)

            # create new task
            for expr in new_expressions:
                task = OptimizePhysicalExpression(expr)
                new_tasks.append(task)

        # mark that the rule has been applied to the logical expression
        self.logical_expression.add_applied_rule(self.rule)

        logger.debug(f"Done applying rule {self.rule} to logical expression {self.logical_expression}")
        logger.debug(f"New tasks: {len(new_tasks)}")
        return new_tasks


class OptimizePhysicalExpression(Task):
    """
    The task to optimize a physical expression and derive its cost.

    This task computes the cost of input groups for the given physical expression (scheduling
    OptimizeGroup tasks if needed), computes the cost of the given expression, and then updates
    the expression's group depending on whether this expression is its `best_physical_expression`
    or in its `pareto_optimal_physical_expressions`.
    """

    def __init__(self, physical_expression: Expression, exploring: bool = False):
        self.physical_expression = physical_expression
        self.exploring = exploring

    def update_best_physical_expression(self, group: Group, policy: Policy) -> Group:
        """
        Update the best physical expression for the given group and policy (if necessary).
        """
        # get the PlanCosts for the current best expression and this physical expression
        best_plan_cost = (
            group.best_physical_expression.plan_cost if group.best_physical_expression is not None else None
        )
        expr_plan_cost = self.physical_expression.plan_cost

        # pre-compute whether or not this physical expression satisfies the policy constraint
        expr_satisfies_constraint = policy.constraint(expr_plan_cost)

        # if we do not have a best physical expression for the group, we set this to be the best expression
        if group.best_physical_expression is None:
            group.best_physical_expression = self.physical_expression
            group.satisfies_constraint = expr_satisfies_constraint

        # if the group currently satisfies the constraint, only update the best physical expression
        # if this expression also satisfies the constraint and is more policy optimal
        elif group.satisfies_constraint and expr_satisfies_constraint and policy.choose(expr_plan_cost, best_plan_cost):
            group.best_physical_expression = self.physical_expression

        # finally, if the group does not satisfy the constraint, update the best physical expression if
        # this expression does satisfy the constraint, or if it is more policy optimal
        elif not group.satisfies_constraint and (
            expr_satisfies_constraint or policy.choose(expr_plan_cost, best_plan_cost)
        ):
            group.best_physical_expression = self.physical_expression
            group.satisfies_constraint = expr_satisfies_constraint

        return group

    def _is_dominated(self, plan_cost: PlanCost, other_plan_cost: PlanCost, policy: Policy):
        """
        Return true if plan_cost is dominated by other_plan_cost and False otherwise.

        If plan costs are perfectly tied on dimensions of interest, other dimensions
        will be used as a tiebreaker.
        """
        # get the dictionary representation of this poicy
        policy_dict = policy.get_dict()

        # get the metrics which matter for this policy
        metrics_of_interest = {metric for metric, weight in policy_dict.items() if weight > 0.0}
        remaining_metrics = {metric for metric, weight in policy_dict.items() if weight == 0.0}

        # corner case: if the two plan costs are perfectly tied on all dimensions of interest,
        # use other dimensions as tiebreaker
        if (
            all([getattr(plan_cost, metric) == getattr(other_plan_cost, metric) for metric in metrics_of_interest])
            and plan_cost.op_estimates.cardinality == other_plan_cost.op_estimates.cardinality
        ):
            for metric in remaining_metrics:
                if metric == "cost" and plan_cost.cost < other_plan_cost.cost:  # noqa: SIM114
                    return False
                elif metric == "time" and plan_cost.time < other_plan_cost.time:  # noqa: SIM114
                    return False
                elif metric == "quality" and plan_cost.quality > other_plan_cost.quality:
                    return False

            # if plan_cost is dominated by other_plan_cost on remaining metrics, return True
            return True

        # normal case: identify whether plan_cost is dominated by other_plan_cost
        cost_dominated = True if policy_dict["cost"] == 0.0 else other_plan_cost.cost <= plan_cost.cost
        time_dominated = True if policy_dict["time"] == 0.0 else other_plan_cost.time <= plan_cost.time
        quality_dominated = True if policy_dict["quality"] == 0.0 else other_plan_cost.quality >= plan_cost.quality
        cardinality_dominated = other_plan_cost.op_estimates.cardinality <= plan_cost.op_estimates.cardinality

        return cost_dominated and time_dominated and quality_dominated and cardinality_dominated

    def _is_pareto_optimal(self, expr_plan_cost: PlanCost, pareto_optimal_physical_expressions: list[Expression], policy: Policy) -> bool:
        """
        Return True if expr_plan_cost is pareto optimal and False otherwise.
        """
        pareto_optimal = True
        for pareto_phys_expr in pareto_optimal_physical_expressions:
            for other_expr_plan_cost, _ in pareto_phys_expr.pareto_optimal_plan_costs:
                if self._is_dominated(expr_plan_cost, other_expr_plan_cost, policy):
                    pareto_optimal = False
                    break

        return pareto_optimal

    def update_pareto_optimal_physical_expressions(self, group: Group, policy: Policy) -> Group:
        """
        Update the pareto optimal physical expressions for the given group and policy (if necessary).
        """
        for pareto_expr_plan_cost, _ in self.physical_expression.pareto_optimal_plan_costs:
            # if the pareto optimal physical expressions are empty, set the pareto optimal
            # physical expressions to be this expression
            if group.pareto_optimal_physical_expressions is None:
                group.pareto_optimal_physical_expressions = [self.physical_expression]

            # otherwise, if this expression is pareto optimal, update the pareto frontier
            elif self._is_pareto_optimal(pareto_expr_plan_cost, group.pareto_optimal_physical_expressions, policy):
                all_physical_expressions = [self.physical_expression] + group.pareto_optimal_physical_expressions

                # compute the pareto optimal set of expressions (or plan costs)
                pareto_optimal_physical_expressions = []
                for idx, expr in enumerate(all_physical_expressions):
                    for plan_cost, _ in expr.pareto_optimal_plan_costs:
                        pareto_optimal = True

                        # check if any other_expr has a plan cost which dominates plan_cost
                        for other_idx, other_expr in enumerate(all_physical_expressions):
                            if idx == other_idx:
                                continue

                            # if plan is dominated by other_expr, set pareto_optimal = False and break
                            for other_plan_cost, _ in other_expr.pareto_optimal_plan_costs:
                                if self._is_dominated(plan_cost, other_plan_cost, policy):
                                    pareto_optimal = False
                                    break

                            # break early if plan_cost is already dominated by another expression's plan_cost
                            if not pareto_optimal:
                                break

                        # add expr to pareto frontier if it has at least one plan cost which is not dominated
                        if pareto_optimal:
                            pareto_optimal_physical_expressions.append(expr)

                            # we can break now because we've identified that this expression has a plan on the pareto frontier
                            break

                # set pareto optimal physical expressions for the group
                group.pareto_optimal_physical_expressions = pareto_optimal_physical_expressions

        return group

    def perform(
        self,
        cost_model: BaseCostModel,
        groups: dict[int, Group],
        policy: Policy,
        context: dict[str, Any] | None = None,
    ) -> list[Task]:
        logger.debug(f"Optimizing physical expression {self.physical_expression}")

        if context is None:
            context = {}

        # get the optimizer strategy (type) from the context
        optimizer_strategy: OptimizationStrategyType = context['optimizer_strategy']

        # return if we've already computed the cost of this physical expression
        if optimizer_strategy.is_pareto() and self.physical_expression.pareto_optimal_plan_costs is not None:
            return []

        if optimizer_strategy.is_not_pareto() and self.physical_expression.plan_cost is not None:
            return []

        # for expressions with an input group, compute the input plan cost(s)
        best_input_plan_cost = PlanCost(cost=0, time=0, quality=1)
        input_plan_costs = [PlanCost(cost=0, time=0, quality=1)]
        if len(self.physical_expression.input_group_ids) > 0:
            # get the input group
            input_group_id = self.physical_expression.input_group_ids[0]  # TODO: need to handle joins
            input_group = groups[input_group_id]

            # compute the input plan cost or list of input plan costs
            new_tasks = []
            if optimizer_strategy.is_not_pareto() and input_group.best_physical_expression is not None:
                # TODO: apply policy constraint here
                best_input_plan_cost = input_group.best_physical_expression.plan_cost

            elif optimizer_strategy.is_pareto() and input_group.pareto_optimal_physical_expressions is not None:
                # TODO: apply policy constraint here
                input_plan_costs = []
                for pareto_physical_expression in input_group.pareto_optimal_physical_expressions:
                    plan_costs = list(map(lambda tup: tup[0], pareto_physical_expression.pareto_optimal_plan_costs))
                    input_plan_costs.extend(plan_costs)

                # NOTE: this list will not necessarily be pareto-optimal, as a plan cost on the pareto frontier of
                # one pareto_optimal_physical_expression might be dominated by the plan cost on another physical
                # expression's pareto frontier; we handle this below by taking the pareto frontier of all_possible_plan_costs
                # de-duplicate equivalent plan costs; we will still reconstruct plans with equivalent cost in optimizer.py
                input_plan_costs = list(set(input_plan_costs))

            else:
                task = OptimizeGroup(input_group_id)
                new_tasks.append(task)

            # if not all input groups have been costed, we need to compute these first and then retry this task
            if len(new_tasks) > 0:
                return [self] + new_tasks

        group = groups[self.physical_expression.group_id]
        if optimizer_strategy.is_pareto():
            # compute all possible plan costs for this physical expression given the pareto optimal input plan costs
            all_possible_plan_costs = []
            for input_plan_cost in input_plan_costs:
                op_plan_cost = cost_model(self.physical_expression.operator, input_plan_cost.op_estimates)

                # compute the total cost for this physical expression by summing its operator's PlanCost
                # with the input groups' total PlanCost; also set the op_estimates for this expression's operator
                full_plan_cost = op_plan_cost + input_plan_cost
                full_plan_cost.op_estimates = op_plan_cost.op_estimates
                all_possible_plan_costs.append((full_plan_cost, input_plan_cost))

            # reduce the set of possible plan costs to the subset which are pareto-optimal
            pareto_optimal_plan_costs = []
            for idx, (plan_cost, input_plan_cost) in enumerate(all_possible_plan_costs):
                pareto_optimal = True

                # check if any other_expr dominates expr
                for other_idx, (other_plan_cost, _) in enumerate(all_possible_plan_costs):
                    if idx == other_idx:
                        continue

                    # if plan is dominated by other_expr, set pareto_optimal = False and break
                    if self._is_dominated(plan_cost, other_plan_cost, policy):
                        pareto_optimal = False
                        break

                # add expr to pareto frontier if it's not dominated
                if pareto_optimal:
                    pareto_optimal_plan_costs.append((plan_cost, input_plan_cost))

            # set the pareto frontier of plan costs which can be obtained by this physical expression
            self.physical_expression.pareto_optimal_plan_costs = pareto_optimal_plan_costs

            # update the group's pareto optimal costs
            group = self.update_pareto_optimal_physical_expressions(group, policy)

        else:
            # otherwise, compute the cost of this operator given the optimal input plan cost
            op_plan_cost = cost_model(self.physical_expression.operator, best_input_plan_cost.op_estimates)

            # compute the total cost for this physical expression by summing its operator's PlanCost
            # with the input groups' total PlanCost; also set the op_estimates for this expression's operator
            full_plan_cost = op_plan_cost + best_input_plan_cost
            full_plan_cost.op_estimates = op_plan_cost.op_estimates
            self.physical_expression.plan_cost = full_plan_cost

            # update the best physical expression for the group
            group = self.update_best_physical_expression(group, policy)

        # set the group's optimized flag to True, store the updated group, and return
        group.optimized = True
        groups[self.physical_expression.group_id] = group

        logger.debug(f"Done optimizing physical expression {self.physical_expression}")
        return []



import logging
from enum import Enum

from palimpzest.core.elements.records import DataRecordCollection
from palimpzest.query.execution.execution_strategy import ExecutionStrategy, SentinelExecutionStrategy
from palimpzest.query.execution.execution_strategy_type import ExecutionStrategyType, SentinelExecutionStrategyType
from palimpzest.query.optimizer.cost_model import CostModel
from palimpzest.query.optimizer.optimizer import Optimizer
from palimpzest.query.optimizer.optimizer_strategy_type import OptimizationStrategyType
from palimpzest.query.processor.config import QueryProcessorConfig
from palimpzest.query.processor.processing_strategy_type import ProcessingStrategyType
from palimpzest.query.processor.query_processor import QueryProcessor
from palimpzest.sets import Dataset, Set
from palimpzest.utils.model_helpers import get_models

logger = logging.getLogger(__name__)


class QueryProcessorFactory:

    @classmethod
    def _convert_to_enum(cls, enum_type: type[Enum], value: str) -> Enum:
        value = value.upper().replace('-', '_')
        try:
            return enum_type[value]
        except KeyError as e:
            raise ValueError(f"Unsupported {enum_type.__name__}: {value}") from e

    @classmethod
    def _normalize_strategies(cls, config: QueryProcessorConfig):
        """
        Convert the string representation of each strategy into its Enum equivalent and throw
        an exception if the conversion fails.
        """
        strategy_types = {
            "processing_strategy": ProcessingStrategyType,
            "execution_strategy": ExecutionStrategyType,
            "sentinel_execution_strategy": SentinelExecutionStrategyType,
            "optimizer_strategy": OptimizationStrategyType,
        }
        for strategy in ["processing_strategy", "execution_strategy", "sentinel_execution_strategy", "optimizer_strategy"]:
            strategy_str = getattr(config, strategy)
            strategy_type = strategy_types[strategy]
            strategy_enum = None
            if strategy_str is not None:
                try:
                    strategy_enum = cls._convert_to_enum(strategy_type, strategy_str)
                except ValueError as e:
                    raise ValueError(f"""Unsupported {strategy}: {strategy_str}.
                                        The supported strategies are: {strategy_type.__members__.keys()}""") from e
            setattr(config, strategy, strategy_enum)
            logger.debug(f"Normalized {strategy}: {strategy_enum}")

        return config

    @classmethod
    def _config_validation_and_normalization(cls, config: QueryProcessorConfig):
        if config.policy is None:
            raise ValueError("Policy is required for optimizer")

        if config.cache:
            raise ValueError("cache=True is not supported yet")
        
        # only one of progress or verbose can be set; we will default to progress=True
        if config.progress and config.verbose:
            print("WARNING: Both `progress` and `verbose` are set to True, but only one can be True at a time; defaulting to `progress=True`")
            config.verbose = False

        # handle "auto" defaults for processing and sentinel execution strategies
        if config.processing_strategy == "auto":
            config.processing_strategy = "no_sentinel" if config.val_datasource is None else "sentinel"

        if config.sentinel_execution_strategy == "auto":
            config.sentinel_execution_strategy = None if config.val_datasource is None else "mab"

        # convert the config values for processing, execution, and optimization strategies to enums
        config = cls._normalize_strategies(config)

        # check that processor uses a supported execution strategy
        if config.execution_strategy not in config.processing_strategy.valid_execution_strategies():
            raise ValueError(f"Unsupported `execution_strategy` {config.execution_strategy} for `processing_strategy` {config.processing_strategy}.")

        # check that validation data is provided for sentinel execution
        if config.val_datasource is None and config.processing_strategy.is_sentinel_strategy():
            raise ValueError("`val_datasource` is required for SENTINEL processing strategies")

        # check that sentinel execution is provided for sentinel processor
        if config.sentinel_execution_strategy is None and config.processing_strategy.is_sentinel_strategy():
            raise ValueError("`sentinel_execution_strategy` is required for SENTINEL processing strategies")

        # get available models
        available_models = getattr(config, 'available_models', [])
        if available_models is None or len(available_models) == 0:
            available_models = get_models(include_vision=True)
        config.available_models = available_models

        return config

    @classmethod
    def _create_optimizer(cls, config: QueryProcessorConfig) -> Optimizer:
        return Optimizer(cost_model=CostModel(), **config.to_dict())

    @classmethod
    def _create_execution_strategy(cls, config: QueryProcessorConfig) -> ExecutionStrategy:
        """
        Creates an execution strategy based on the configuration.
        """
        execution_strategy_cls = config.execution_strategy.value
        return execution_strategy_cls(**config.to_dict())

    @classmethod
    def _create_sentinel_execution_strategy(cls, config: QueryProcessorConfig) -> SentinelExecutionStrategy:
        """
        Creates an execution strategy based on the configuration.
        """
        if config.sentinel_execution_strategy is None:
            return None

        sentinel_execution_strategy_cls = config.sentinel_execution_strategy.value
        return sentinel_execution_strategy_cls(**config.to_dict())

    @classmethod
    def create_processor(
        cls,
        dataset: Set,
        config: QueryProcessorConfig | None = None,
        **kwargs
    ) -> QueryProcessor:
        """
        Creates a QueryProcessor with specified processing and execution strategies.

        Args:
            dataset: The dataset to process
            config: The user-provided QueryProcessorConfig; if it is None, the default config will be used
            kwargs: Additional keyword arguments to pass to the QueryProcessorConfig
        """
        if config is None:
            config = QueryProcessorConfig()

        # apply any additional keyword arguments to the config and validate its contents
        config.update(**kwargs)
        config = cls._config_validation_and_normalization(config)

        # create the optimizer, execution strateg(ies), and processor
        optimizer = cls._create_optimizer(config)
        config.execution_strategy = cls._create_execution_strategy(config)
        config.sentinel_execution_strategy = cls._create_sentinel_execution_strategy(config)
        processor_cls = config.processing_strategy.value
        processor = processor_cls(dataset, optimizer, **config.to_dict())

        return processor

    @classmethod
    def create_and_run_processor(cls, dataset: Dataset, config: QueryProcessorConfig | None = None, **kwargs) -> DataRecordCollection:
        # TODO(Jun): Consider to use cache here.
        logger.info(f"Creating processor for dataset: {dataset}")
        processor = cls.create_processor(dataset=dataset, config=config, **kwargs)
        logger.info(f"Created processor: {processor}")

        return processor.execute()



import logging

from palimpzest.constants import Cardinality
from palimpzest.core.data.datareaders import DataReader
from palimpzest.policy import (
    MaxQuality,
    MaxQualityAtFixedCost,
    MaxQualityAtFixedTime,
    MinCost,
    MinCostAtFixedQuality,
    MinTime,
    MinTimeAtFixedQuality,
    PlanCost,
    Policy,
)
from palimpzest.query.processor.config import QueryProcessorConfig
from palimpzest.sets import Dataset

# Initialize the root logger
logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    # constants
    "Cardinality",
    # core
    "DataReader",
    # policy
    "MaxQuality",
    "MaxQualityAtFixedCost",
    "MaxQualityAtFixedTime",
    "MinCost",
    "MinCostAtFixedQuality",
    "MinTime",
    "MinTimeAtFixedQuality",
    "PlanCost",
    "Policy",
    # query
    "QueryProcessorConfig",
    # sets
    "Dataset",
]



from __future__ import annotations


class Field:
    """
    A Field is defined by its description and its type. The Field class is subclassed to specify
    that values of the subclass should belong to a specific type.

    For example, if you wanted to define Fields relevant to indexing research papers, you could define a field
    representing the title of a paper, the year it was published, and the journal it was published in:

    ```python
    paper_title = Field(desc="The title of a scientific paper")
    paper_year = Field(desc="The year the paper was published")
    paper_journal = Field(desc="The name of the journal that published the paper")
    ```
    """
    is_image_field = False

    def __init__(self, desc: str = "") -> None:
        self._desc = desc
        self.type = None

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(desc={self._desc})"

    def __hash__(self) -> int:
        return hash(self._desc + self.__class__.__name__)

    def __eq__(self, other) -> bool:
        return self._desc == other._desc and self.__class__ == other.__class__

    @property
    def desc(self) -> str:
        return self._desc

    def json_schema(self) -> dict:
        return {"description": self._desc, "type": str(self.type)}


class BooleanField(Field):
    """A BooleanField is a Field that is True or False."""

    def __init__(self, desc: str):
        super().__init__(desc=desc)
        self.type = bool


class BytesField(Field):
    """A BytesField is a Field that is definitely an array of bytes."""

    def __init__(self, desc: str):
        super().__init__(desc=desc)
        self.type = bytes

    def json_schema(self) -> dict[str, str]:
        return {
            "description": self._desc,
            "type": str(self.type),
            "contentEncoding": "base64",
            "contentMediaType": "application/octet-stream",
        }


class CallableField(Field):
    """A CallableField is a Field that stores a function."""

    def __init__(self, desc: str):
        super().__init__(desc=desc)
        self.type = type(lambda x: x)


class FloatField(Field):
    """A FloatField is a Field that is definitely an integer or a float."""

    def __init__(self, desc: str):
        super().__init__(desc=desc)
        self.type = float


class IntField(Field):
    """An IntField is a Field that is definitely an integer or a float."""

    def __init__(self, desc: str):
        super().__init__(desc=desc)
        self.type = int


class NumericField(Field):
    """A NumericField is a Field that is definitely a number."""

    def __init__(self, desc: str):
        super().__init__(desc=desc)
        self.type = int | float


class StringField(Field):
    """A StringField is a Field that is definitely a string of text."""

    def __init__(self, desc: str):
        super().__init__(desc=desc)
        self.type = str


class ImageFilepathField(StringField):
    """An ImageFilepathField is a StringField that contains the filepath to an image."""
    is_image_field = True

    def __init__(self, desc: str):
        super().__init__(desc=desc)


class ImageURLField(StringField):
    """An ImageURLField is a StringField that contains the publicly accessible URL for an image."""
    is_image_field = True

    def __init__(self, desc: str):
        super().__init__(desc=desc)


class ImageBase64Field(BytesField):
    """An ImageBase64Field is a BytesField that contains a base64 encoded image."""
    is_image_field = True

    def __init__(self, desc: str):
        super().__init__(desc=desc)


### fields which are metaclasses that produce other field types ###
class ListField(Field):
    """A field representing a list of elements of specified types, with full list functionality."""

    def __new__(cls, element_type: Field, desc: str | None = None):
        attrs = {
            "element_type": element_type,
            "is_image_field": element_type.is_image_field,
            "type": list,
            "_desc": desc,
        }

        return type(f"List[{element_type.__name__}]", (Field,), attrs)



"""
This file collects a sample of useful UDFs to convert schemata.
"""

import io
from datetime import datetime

import pandas as pd
import requests

from palimpzest.constants import MAX_ROWS


def url_to_file(candidate: dict):
    """Function used to convert a DataRecord instance of URL to a File DataRecord."""
    url = candidate["url"]
    filename = url.split("/")[-1]
    timestamp = datetime.now().isoformat()
    try:
        contents = requests.get(url).content
    except Exception as e:
        print(f"Error fetching URL {url}: {e}")
        contents = b""

    return {"filename": filename, "timestamp": timestamp, "contents": contents}


def file_to_xls(candidate: dict):
    """Function used to convert a DataRecord instance of File to a XLSFile DataRecord."""
    xls = pd.ExcelFile(io.BytesIO(candidate["contents"]), engine="openpyxl")
    return {"number_sheets": len(xls.sheet_names), "sheet_names": xls.sheet_names}


def xls_to_tables(candidate: dict):
    """Function used to convert a DataRecord instance of XLSFile to a Table DataRecord."""
    xls_bytes = candidate["contents"]
    sheet_names = candidate["sheet_names"]

    records = []
    for sheet_name in sheet_names:
        dataframe = pd.read_excel(io.BytesIO(xls_bytes), sheet_name=sheet_name, engine="openpyxl")

        # TODO extend number of rows with dynamic sizing of context length
        # construct data record
        record = {}
        rows = []
        for row in dataframe.values[:100]:
            row_record = [str(x) for x in row]
            rows += [row_record]
        record["rows"] = rows[:MAX_ROWS]
        record["filename"] = candidate["filename"]
        record["header"] = dataframe.columns.values.tolist()
        record["name"] = candidate["filename"].split("/")[-1] + "_" + sheet_name
        records.append(record)

    return records
